/*
 *  Copyright (C) 2002-2005 The SSI Project. All rights reserved.
 *  http://ssi.is.s.u-tokyo.ac.jp
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY.  See the COPYING file for the precise
 *  terms and conditions for the use of this software.
 */
#include "libfftss.h"

void fftss_execute(fftss_plan p)
{
  double *b0, *b1, *b2, *b3;
  long i;

  b0 = p->in;
  b1 = p->out;

  if (p->flags & FFTSS_INOUT) {

    if (p->logn2 < 2) {
      if (p->logn2 == 1) {
	double xr0, xi0, xr1, xi1;
	xr0 = b0[0] + b0[2];	xi0 = b0[1] + b0[3];
	xr1 = b0[0] - b0[2];	xi1 = b0[1] - b0[3];
	b0[0] = xr0; b0[1] = xi0; b0[2] = xr1; b0[3] = xi1;
      }
      return;
    }

    if (p->stages & 1) {
      b1 = p->in;
      b2 = p->out;
    } else {
      b1 = p->out;
      b2 = p->in;
    }
    
  } else {
    if (p->logn2 < 2) {
      if (p->logn2 == 0) {
	b1[0] = b0[0];
	b1[1] = b0[1];
      } else {
	b1[0] = b0[0] + b0[2];	b1[1] = b0[1] + b0[3];
	b1[2] = b0[0] - b0[2];	b1[3] = b0[1] - b0[3];
      }
      return;
    }

    if (p->stages & 1) {
      b1 = p->out;
      if (p->flags & FFTSS_PRESERVE_INPUT) b2 = p->work;
      else b2 = p->in;
    } else {
      if (p->flags & FFTSS_PRESERVE_INPUT) b1 = p->work;
      else b1 = p->in;
      b2 = p->out;
    }
  }
   
  p->k[0].kern(b0, b1, p->w, p->k[0].bsize, p->k[0].blocks);

  for (i = 1; i < p->stages; i++) {
    p->k[i].kern(b1, b2, p->w, p->k[i].bsize, p->k[i].blocks);
    b3 = b2; b2 = b1; b1 = b3;
  }

}
