/*
 *  Copyright (C) 2002-2005 The SSI Project. All rights reserved.
 *  http://ssi.is.s.u-tokyo.ac.jp
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY.  See the COPYING file for the precise
 *  terms and conditions for the use of this software.
 */
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include "libfftss.h"

#if defined(HAVE_MEMALIGN) || defined(HAVE_POSIX_MEMALIGN)

void *fftss_malloc(long size)
{
#if defined(HAVE_MEMALIGN)
  return memalign(16, size);
#elif defined(HAVE_POSIX_MEMALIGN)
  return posix_memalign(16, size);
#endif
}

void fftss_free(void *p)
{
  free(p);
  return;
}

#else

typedef struct _malloc_address {
  struct _malloc_address *next, *prev;
  void *address, *real_address;
} malloc_address;

malloc_address malloc_address_top = 
  {&malloc_address_top, &malloc_address_top, NULL, NULL};

void *fftss_malloc(long size)
{
  union {
    void *ptr;
    unsigned long l;
  } addr;
  malloc_address *ma;

  addr.ptr = malloc(size);
  if (addr.l & 0xf) {
    free(addr.ptr);
	addr.ptr = malloc(size + 16);
    ma = malloc(sizeof(malloc_address));
    ma->next = &malloc_address_top;
	ma->prev = malloc_address_top.prev;
	ma->prev->next = ma;
	ma->next->prev = ma;
	ma->real_address = addr.ptr;
	addr.l += 0xf;
	addr.l &= ~0xf;
	ma->address = addr.ptr;
  }

/*    printf("warning: allocated memory is not aligned.\n");  */
  
  return addr.ptr;
}

void fftss_free(void *p)
{
  malloc_address *ma;

  ma = malloc_address_top.next;
  while (ma->address) {
    if (ma->address == p) break;
	ma = ma->next;
  }
  if (ma->address) {
    ma->next->prev = ma->prev;
	ma->prev->next = ma->next;
    free(ma->real_address);
	free(ma);
	return;
  }
  free(p);
  return;
}

#endif

